/*------------------------------------------------------------------
Datei: twi_driver.c 

Programm: Funktionen fr TWI: Master-Transmitter-Receiver Mode 

	m_t(slave, startadr, len, *buf, buf_offset):
			slave: Empfnger, siehe ukt.h
			startadr: Startadresse im Empfnger
			len: Anzahl der zu sendenten Bytes
			*buf: Datenquelle - Vektor
			buf_offset: ab welcher Stelle im Buffer soll gesendet werden
									1.Stelle = 0 , usw.
			Rckgabewert: Anzahl d. gesendten Bytes , 0 = Fehler

	m_r(slave, startadr, len,  *buf, buf_offset):
			slave: auszulesender Slave, siehe ukt.h
			startadr: ab welcher Adresse soll gelesen werden
			len: wieviele Bytes lesen		
			*buf: Datenziel - Vektor
			buf_offset: ab welcher Stelle im Buffer soll gespeichert werden
									1.Stelle = 0 , usw.
			Rckgabewert: Anzahl d. gelesenen Bytes , 0 = Fehler

Autor:bay	 Datum: 11.08.06

gendert:17.12.06 twi_driver_5.c auf twi_drv.c
24.01.07 twi_drv.c auf twi_driver.c																		 
-------------------------------------------------------------------*/
#include <avr/io.h>
#include <C:\WinAVR-20100110\avr\include\compat\twi.h>	//home
#include "ukt.h"

/* Schreibe auf Slave mit Startadresse len Bytes aus Datenbuffer 0 + Offset */

uchar m_t(uchar slave, uchar startadr, signed char len, uchar *buf,uchar buf_offset)
{
	uchar slave_adr;
	uchar rv = 0; // return value: gesendete Bytes, 0 = Fehler

	buf += buf_offset;	//ab welcher Stelle im Buffer soll gesendet werden

	switch(slave)	//welcher Slave soll angesprochen werden?
	{
		case 0: slave_adr = w_adr_8574_0;
						break;
//		case 1: slave_adr = w_adr_8574_1;
//						break;
		case 2: slave_adr = w_adr_1064_0;
						break;
		case 3: slave_adr = w_adr_8583_0;
						break;
		case 4: slave_adr = w_adr_9801_0;
						break;
	}

	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);		//TWI Start
	while(!(TWCR & (1<<TWINT)));									//warte auf TWINT Flag
	if((TWSR & 0xF8) != start) goto twi_error; 		//prfe TWI Status 0x08

	TWDR = slave_adr;															// Slave Adresse + Write
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)));
	if((TWSR & 0xF8) != mt_sla_ack) goto twi_error;	//prfe TWI Status 0x18

	if((slave == 0) ||(slave == 1)) goto weiter;

	TWDR = startadr; 						// Startadresse fr erstes Byte
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)));
	if((TWSR & 0xF8) != mt_data_ack) goto twi_error;	//prfe TWI Status 0x28
	rv++;

	weiter:

	for(;len>0;len--)
	{
		TWDR = *buf++; 									// len Datenbytes senden
		TWCR = (1<<TWINT) | (1<<TWEN);
		while(!(TWCR & (1<<TWINT)));
		if((TWSR & 0xF8) != mt_data_ack) goto twi_error;	//prfe TWI Status 0x28
		rv++;
	}

	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO); 				// TWI Stopp
	
	out:
	return rv;	
	twi_error: rv = 0;
	goto out;
}

/* Lese aus Slave len Bytes nach Datenbuffer 0 + Offset */

uchar m_r(uchar slave, uchar startadr, signed char len, uchar *buf,uchar buf_offset)
{
	uchar slave_adr_r, slave_adr_w, twcr, twsr;

	uchar rv = 0; // return value: empfangene Bytes, 0 = Fehler

	buf += buf_offset;	//nach welcher Stelle im Buffer soll geschrieben werden

	switch(slave)	//welcher Slave soll angesprochen werden?
	{
		case 0: slave_adr_r = r_adr_8574_0;
						slave_adr_w =	w_adr_8574_0;						
						break;

//		case 1: slave_adr_r = r_adr_8574_1;	
//						slave_adr_w =	w_adr_8574_1;
//						break;

		case 3: slave_adr_r = r_adr_8583_0;
						slave_adr_w = w_adr_8583_0;	
						break;

		case 4: slave_adr_r = r_adr_9801_0;
						slave_adr_w = w_adr_9801_0;	
						break;
	}

	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);		//TWI Start
	while(!(TWCR & (1<<TWINT)));									//warte bis bertragung fertig 
	if((TWSR & 0xF8) != start) goto twi_error; 		//prfe TWI Status 0x08

	if((slave == 0) ||(slave == 1)) goto weiter_1;

	TWDR = slave_adr_w;															// Slave Adresse + Write
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)));
	if((TWSR & 0xF8) != mt_sla_ack) goto twi_error;	//prfe TWI Status 0x18

	TWDR = startadr; 													// Startadresse fr erstes Byte
	TWCR = (1<<TWINT) | (1<<TWEN);
	while(!(TWCR & (1<<TWINT)));
	if((TWSR & 0xF8) != mt_data_ack) goto twi_error;	//prfe TWI Status 0x28

	TWCR = (1<<TWINT) | (1<<TWSTA) | (1<<TWEN);		//TWI Rep.Start
	while(!(TWCR & (1<<TWINT)));									//warte bis bertragung fertig 
	if((TWSR & 0xF8) != restart) goto twi_error; 		//prfe TWI Status 0x10

weiter_1:

	TWDR = slave_adr_r;							//MR Mode							
	TWCR = (1<<TWINT) | (1<<TWEN);	//SLA + R senden, ACK empfangen
	while(!(TWCR & (1<<TWINT)));		
	if((TWSR & 0xF8) != mr_sla_ack) goto twi_error;	//prfe TWI Status 0x40

	twcr = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);

	for(; len > 0; len--)
	{
		if(len == 1) twcr = (1<<TWINT) | (1<<TWEN);	//Sende NACK

		TWCR = twcr;	//starte Einlesen

		while(!(TWCR & (1<<TWINT)));
		
		twsr = (TWSR & 0xF8);	//Statusregister filtern			

		switch(twsr)
		{
			case mr_data_nack: len = 0;	//prfe TWI Status 0x58

			case mr_data_ack: //prfe TWI Status 0x50

					*buf++ = TWDR; 	// empfangenes Datenbyte speichern
					rv++;
					break;

			default: goto twi_error;
		}
	}

	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);		// TWI Stopp
	
	out:
	return rv;	
	twi_error: rv = 0;
	goto out;
}


